from typing import Callable, Coroutine

from flask import Flask, Blueprint

from ..abc import BaseApi, BaseRouter

__all__ = ("FlaskAPI", "FlaskRouter")


class FlaskRouter(BaseRouter):
    def __init__(self, name: str, **kwargs):
        self.blueprint = Blueprint(name.lower(), name, **kwargs)

    def add_endpoint_method(self, coro: Callable[..., Coroutine], endpoint: str, method: str, **kwargs):
        self.blueprint.route(endpoint, methods=[method], **kwargs)(coro)


class FlaskAPI(BaseApi):
    def __init__(self, host: str, port: int, **kwargs):
        self.host = host
        self.port = port
        self.debug = kwargs.pop("debug", False)

        self.app: Flask = Flask(kwargs.pop("import_name", None) or "interactions_restful", **kwargs)

    def add_endpoint_method(self, coro: Callable[..., Coroutine], endpoint: str, method: str, **kwargs):
        self.app.route(endpoint, methods=[method], **kwargs)(coro)

    @staticmethod
    def create_router(**kwargs):
        return FlaskRouter(kwargs.pop("name"), **kwargs)

    def add_router(self, router: FlaskRouter):
        self.app.register_blueprint(router.blueprint)

    async def run(self):
        self.app.run(self.host, self.port, self.debug)



import asyncio
from threading import Thread
from inspect import getmembers
from typing import Type, TypeVar

from interactions import Client

from .abc import BaseApi

__all__ = ("APIClient", "setup")
API_TYPE = TypeVar("API_TYPE", bound=BaseApi)


class APIClient:
    bot: Client

    def __init__(self, bot: Client, api: Type[API_TYPE], host: str, port: int, **kwargs) -> None:
        self.bot = bot
        self.bot.api = self

        self.client: API_TYPE = api(host, port, **kwargs)

        self.bot.async_startup_tasks.append((self.run, (), {}))

    async def run(self):
        self._register_extension_routes()

        thread = Thread(target=self._run_api_thread_worker)
        thread.start()

    def _run_api_thread_worker(self):
        asyncio.run(self.client.run())

    def _register_extension_routes(self):
        for extension in self.bot.ext.values():
            has_methods = False
            for _, coro in getmembers(extension):
                if getattr(coro, "__api__", None) is None:
                    continue

                has_methods = True

                if getattr(extension, "router", None) is None:
                    extension.router = self.client.create_router(name=extension.name)
                data: dict = coro.__api__
                extension.router.add_endpoint_method(coro, data["endpoint"], data["method"], **data["kwargs"])

            if has_methods:
                self.client.add_router(extension.router)


def setup(client: Client, api: Type[API_TYPE], host: str, port: int, **kwargs) -> APIClient:
    return APIClient(client, api, host, port, **kwargs)


import typing

from interactions import Absent, MISSING

__all__ = ("route",)


def route(method: str, endpoint: Absent[str] = MISSING, **kwargs):
    def wrapper(coro: typing.Callable[..., typing.Coroutine]):
        coro.__api__ = {
            "method": method,
            "endpoint": endpoint or coro.__name__,
            "kwargs": kwargs
        }

        return coro
    return wrapper



from typing import Callable, Coroutine

from fastapi import FastAPI as _FastAPI, APIRouter
from uvicorn import Config, Server

from ..abc import BaseApi, BaseRouter

__all__ = ("FastAPI", )


class FastApiRouter(BaseRouter):
    def __init__(self, **kwargs):
        kwargs.pop("name")  # neccessary only for flask
        self.api_router = APIRouter(**kwargs)

    def add_endpoint_method(self, coro: Callable[..., Coroutine], endpoint: str, method: str, **kwargs):
        self.api_router.add_api_route(
            endpoint,
            coro,
            methods=[method],
            **kwargs
        )


class FastAPI(BaseApi):
    def __init__(self, host: str, port: int, **kwargs):
        self.host = host
        self.port = port

        self.app = _FastAPI(**kwargs)
        self._config = Config(self.app, host=self.host, port=self.port)

    def add_endpoint_method(self, coro: Callable[..., Coroutine], endpoint: str, method: str, **kwargs):
        self.app.router.add_api_route(
            endpoint,
            coro,
            methods=[method],
            **kwargs
        )

    @staticmethod
    def create_router(**kwargs) -> BaseRouter:
        return FastApiRouter(**kwargs)

    def add_router(self, router: FastApiRouter):
        self.app.include_router(router.api_router)

    async def run(self):
        server = Server(config=self._config)
        await server.serve()



from abc import ABC, abstractmethod
from typing import Callable, Coroutine

__all__ = ("BaseApi", "BaseRouter")


class BaseRouter(ABC):
    @abstractmethod
    def add_endpoint_method(self, coro: Callable[..., Coroutine], endpoint: str, method: str, **kwargs):
        pass


class BaseApi(ABC):
    @abstractmethod
    def __init__(self, host: str, port: int, **kwargs):
        pass

    @abstractmethod
    def add_endpoint_method(self, coro: Callable[..., Coroutine], endpoint: str, method: str, **kwargs):
        pass

    @staticmethod
    @abstractmethod
    def create_router(**kwargs) -> BaseRouter:
        pass

    def add_router(self, router: BaseRouter):
        pass

    @abstractmethod
    async def run(self):
        pass



from codecs import open
from pathlib import Path

import tomli
from setuptools import find_packages, setup

with open("pyproject.toml", "rb") as f:
    pyproject = tomli.load(f)

extras_require = {
    "flask": ["flask>=2.0.0,<=3.0.0"],
    "fastapi": ["fastapi>=0.92.0", "uvicorn>=0.20.0"]
}

setup(
    name=pyproject["tool"]["poetry"]["name"],
    version=pyproject["tool"]["poetry"]["version"],
    author="Damego",
    author_email="damego.dev@gmail.com",
    description=pyproject["tool"]["poetry"]["description"],
    include_package_data=True,
    install_requires=["discord-py-interactions>=5.0.0,<6.0.0"],
    extras_require=extras_require,
    license=pyproject["tool"]["poetry"]["license"],
    long_description=(Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://github.com/interactions-py/restful",
    packages=find_packages(),
    python_requires=">=3.10",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
    ]
)
